// Utility functions

#ifdef WIN32
	#include <windows.h>
	#include <winsock.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "xc.h"
#include "xctypes.h"
#include "xcvars.h"

int do_xinst(xinst *x);

// "Private" data read/write functions

int log2(_int64 m)
{
	int j;
#ifdef WIN32
	_int64 bm = 0x8000000000000000;
#else
	_int64 bm = 0x8000000000000000ULL;
#endif

	for (j = 63; j >= 0; --j)
	{
		if (m & bm)
			return j;
		bm >>= 1;
	}

	return -1;
}

void read_int(unsigned char *str, int sn, _int64 *i)
{
	unsigned _int64 u = 0;
	int k, osz = tsz[sn];

	for (k = 0; k < osz; ++k)
		u += (_int64)*str++ << (8 * k);

	if (u > msb[sn])
		*i = -(signed _int64)(imax[sn] - u + 1);
	else
		*i = u;
}

void write_int(unsigned char *str, int sn, _int64 i)
{
	unsigned _int64 u;
	int k, osz = tsz[sn];

	if (i < 0)
		u = imax[sn] + i + 1;
	else
		u = i;

	for (k = 0; k < osz; ++k)
	{
		*str++ = (char)(u & 255);
		u >>= 8;
	}
}

void print_bin(char *str, int len)
{
	int j, k;
	unsigned char ch;

	printf("Binary:");
	for (j = len-1; j >= 0; --j)
	{
		putchar(' ');
		ch = str[j];
		for (k = 0; k < 8; ++k)
		{
			printf("%d", !!(ch & 128));
			ch <<= 1;
		}
	}
	putchar('\n');
}

void print_float(_int64 m, _int32 e, int ln)
{
	printf("Float: %g\n", m * pow(2, e - mantissa[ln]));
}

int read_float(unsigned char *str, int ln, _int64 *m, _int32 *e)
{
	int len = tsz[ln], j, sign = 1;
	int mn = mantissa[ln], en = exponent[ln];
	char mode = 0;

	*m = *e = 0;

	for (j = 0; j < len; ++j)
	{
		switch (mode)
		{
		case 0:
			if (mn >= 8)
			{
				*m += (_int64)*str++ << (mantissa[ln] - mn);
				mn -= 8;
			}
			else
			{
				*m += (_int64)(*str & (255 >> (8 - mn))) <<
					(mantissa[ln] - mn);
				*e = (int)(*str++ & (255 << mn)) >> mn;
				en -= 8 - mn;
				mode = 1;
			}
			break;
		case 1:
			if (en < 8)
			{
				if (*str & 128)
					sign = -1;
				*e += (*str++ & 127) << (exponent[ln] - en);
			}
			else
			{
				*e += *str++ << en;
				en -= 8;
			}
		}
	}

	if (*m || *e)
	{
		*m = sign * (*m + ((_int64)1 << mantissa[ln]));
		*e -= bias[ln];
	}
	return 0;
}

int read_float2(_int64 val, int ln, _int64 *m, _int32 *e)
{
	int len = tsz[ln], j, sign = 1;
	int mn = mantissa[ln], en = exponent[ln];
	char mode = 0;

	*m = *e = 0;

	for (j = 0; j < len; ++j)
	{
		switch (mode)
		{
		case 0:
			if (mn >= 8)
			{
				*m += (val & 255) << (mantissa[ln] - mn);
				val >>= 8;
				mn -= 8;
			}
			else
			{
				*m += (_int64)(val & (255 >> (8 - mn))) <<
					(mantissa[ln] - mn);
				*e = (int)((val & 255) & (255 << mn)) >> mn;
				val >>= 8;
				en -= 8 - mn;
				mode = 1;
			}
			break;
		case 1:
			if (en < 8)
			{
				if (val & 128)
					sign = -1;
				*e += (int)((val & 127) << (exponent[ln] - en));
				val >>= 8;
			}
			else
			{
				*e += (int)((val & 255) << en);
				val >>= 8;
				en -= 8;
			}
		}
	}

	if (*m || *e)
	{
		*m = sign * (*m + ((_int64)1 << mantissa[ln]));
		*e -= bias[ln];
	}
	return 0;
}

void write_float(unsigned char *otp, _int64 m, _int32 e, int ln)
{
	int j, len2 = tsz[ln], mn = mantissa[ln], en = exponent[ln];
#ifdef WIN32
	_int64 bm = 0x8000000000000000;
#else
	_int64 bm = 0x8000000000000000ULL;
#endif
	unsigned char mode = 0, sign = 0, *csp = otp;

	if (m < 0)
	{
		m = -m;
		sign = 128;
	}

	for (j = 63; j >= 0; --j)
	{
		if (m & bm)
			break;
		bm >>= 1;
	}

	if (j < 0)
		e = 0;
	else if (j == mn)
		e += bias[ln];
	else if (j < mn)
	{
		m <<= mn - j;
		e += bias[ln] - (mn - j);
	}
	else
	{
		m >>= j - mn;
		e += bias[ln] + (j - mn);
	}
	m &= mmask[ln];

	if (e < 0)
	{
		e = 0;
		m = 0;
	}
	else if (e >= fullexp[ln])
	{
		e = fullexp[ln];
		m = 0;
	}

	for (j = 0; j < len2; ++j)
	{
		switch (mode)
		{
		case 0:
			if (mn >= 8)
			{
				*otp++ = (char)(m & 255);
				m >>= 8;
				mn -= 8;
			}
			else
			{
				*otp++ = (char)(((e & (255 >> mn)) << mn) | m);
				en -= 8 - mn;
				e >>= 8 - mn;
				mode = 1;
			}
			break;
		case 1:
			if (en >= 8)
			{
				*otp++ = e & 255;
				e >>= 8;
				en -= 8;
			}
			else
			{
				*otp++ = sign | e;
			}
			break;
		}
	}
}

_int64 write_float2(_int64 m, _int32 e, int ln)
{
	int j, len2 = tsz[ln], mn = mantissa[ln], en = exponent[ln], s = 0;
#ifdef WIN32
	_int64 bm = 0x8000000000000000;
#else
	_int64 bm = 0x8000000000000000ULL;
#endif
	unsigned char mode = 0, sign = 0;
	_int64 val = 0;

	if (m < 0)
	{
		m = -m;
		sign = 128;
	}

	for (j = 63; j >= 0; --j)
	{
		if (m & bm)
			break;
		bm >>= 1;
	}

	if (j < 0)
		e = 0;
	else if (j == mn)
		e += bias[ln];
	else if (j < mn)
	{
		m <<= mn - j;
		e += bias[ln] - (mn - j);
	}
	else
	{
		m >>= j - mn;
		e += bias[ln] + (j - mn);
	}
	m &= mmask[ln];

	if (e < 0)
	{
		e = 0;
		m = 0;
	}
	else if (e >= fullexp[ln])
	{
		e = fullexp[ln];
		m = 0;
	}

	for (j = 0; j < len2; ++j)
	{
		switch (mode)
		{
		case 0:
			if (mn >= 8)
			{
				val |= (_int64)(m & 255) << (8 * s++);
				m >>= 8;
				mn -= 8;
			}
			else
			{
				val |= (_int64)((((e & (255 >> mn)) << mn) | m) & 255) << (8 * s++);
				en -= 8 - mn;
				e >>= 8 - mn;
				mode = 1;
			}
			break;
		case 1:
			if (en >= 8)
			{
				val |= (_int64)(e & 255) << (8 * s++);
				e >>= 8;
				en -= 8;
			}
			else
			{
				val |= (_int64)((sign | e) & 255) << (8 * s++);
			}
			break;
		}
	}

	return val;
}

// Functions for host funcs to access VM features

void *_xc_malloc(xinst *c_inst_p, int size)
{
	int n;

	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return NULL;
	}

	if ((n = __xc_malloc(c_inst_p, size)) == -1)
		return NULL;

	return &c_inst_p->mem[n];
}

int _xc_free(xinst *c_inst_p, void *ptr)
{
	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return 0;
	}

	__xc_free(c_inst_p, (int)ptr - (int)c_inst_p->mem);
	return !xc_get_errno();
}

int _xc_block_size(xinst *c_inst_p, void *ptr)
{
	memblock *mb;
	_int64 n;
	xinst *_xinst = c_inst_p;

	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return 0;
	}
	
	n = xc_ptr_out(ptr);
	for (mb = c_inst_p->mb_base; mb; mb = mb->next)
		if (n == mb->min)
			return mb->max - mb->min + 1;

	return -1;
}

int _xc_ptr_ok(xinst *c_inst_p, void *ptr)
{
	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return 0;
	}

	return (int)ptr >= (int)c_inst_p->mem && (int)ptr < (int)&c_inst_p->mem[c_inst_p->memsz];
}

// Argument _int64->proper format functions

char xc_char_in(_int64 val)
{
#ifdef WEIRD
	return (char)(val & 255);
#else
	return (char)val;
#endif
}

_int16 xc_int16_in(_int64 val)
{
#ifdef WEIRD
	_int16 ret = 0;
	int j;

	for (j = 1; j >= 0; --j)
	{
		ret += (_int16)((val & 255) << (8 * (1-j)));
		val >>= 8;
	}

	return ret;
#else
	return *(_int16 *)&val;
#endif
}

_int32 xc_int32_in(_int64 val)
{
#ifdef WEIRD
	_int32 ret = 0;
	int j;

	for (j = 3; j >= 0; --j)
	{
		ret += (_int32)((val & 255) << (8 * (3-j)));
		val >>= 8;
	}

	return ret;
#else
	return *(_int32 *)&val;
#endif
}

_int64 xc_int64_in(_int64 val)
{
	return val;
}

_float32 xc_float32_in(_int64 val)
{
//#ifdef WEIRD
	_int32 e;
	_int64 m;

	read_float2(val, SIZE_32, &m, &e);
	return (_float32)m * (_float32)pow(2, e - mantissa[SIZE_32]);
/*#else
	log_printf("xc_float32_in: %g\n", *(_float32 *)&val);

	return *(_float32 *)&val;
#endif*/
}

_float64 xc_float64_in(_int64 val)
{
//#ifdef WEIRD
	_int32 e;
	_int64 m;

	read_float2(val, SIZE_64, &m, &e);
	return (_float64)m * (_float64)pow(2, e - mantissa[SIZE_64]);
/*#else
	return *(_float64 *)&val;
#endif*/
}

void *_xc_ptr_in(xinst *c_inst_p, _int64 val, int size)
{
//#ifdef WEIRD
	int j;
//#endif
	_int32 ptr = 0;
	void log_printf(const char *, ...);

	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return NULL;
	}

	if (!val)
		return NULL;

/*	ptr = *(_int32 *)&val;
	printf("Second = %d\n", ptr);

	ptr = 0;*/
	for (j = 3; j >= 0; --j)
	{
		ptr += (_int32)((val & 255) << (8 * (3-j)));
		val >>= 8;
	}
//	printf("First = %d\n", ptr);

	if (ptr < 0 || (ptr + size) > (int)c_inst_p->memsz)
	{
		xc_set_errno(XC_ERR_SEGFAULT);
		return NULL;
	}

	return &c_inst_p->mem[ptr];
}

char *_xc_string_in(xinst *c_inst_p, _int64 val)
{
	int j, sz;
	_int32 ptr = 0;

	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return NULL;
	}

	if (!val)
		return NULL;

#ifdef WEIRD
	for (j = 3; j >= 0; --j)
	{
		ptr += (_int32)((val & 255) << (8 * (3-j)));
		val >>= 8;
	}
#else
	ptr = *(_int32 *)&val;
#endif

	if (ptr < 0 || ptr >= (int)c_inst_p->memsz)
	{
		xc_set_errno(XC_ERR_SEGFAULT);
		return NULL;
	}

	sz = c_inst_p->memsz;
	for (j = ptr; j < sz; ++j)
		if (!c_inst_p->mem[j])
			return &c_inst_p->mem[ptr];

	xc_set_errno(XC_ERR_SEGFAULT);
	return NULL;
}

_int32 _xc_fptr_in(xinst *c_inst_p, _int64 val)
{
	_int32 n = xc_int32_in(val);

	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return 0;
	}

	if (n < 0 || n > c_inst_p->code_end - c_inst_p->code)
	{
		xc_set_errno(XC_ERR_SEGFAULT);
		return 0;
	}

	return n;
}

// Return value format->_int64 converters

_int64 xc_char_out(char val)
{
	return (_int64)val;
}

_int64 xc_int16_out(_int16 val)
{
#ifdef WEIRD
	int j;
	_int64 ret = 0;

	for (j = 1; j >= 0; --j)
	{
		ret += (val & 255) << (8 * (1-j));
		val >>= 8;
	}

	return ret;
#else
	unsigned char buf[8];

	memcpy(buf, &val, 2);
	memset(buf+2, 0, 6);
	return *(_int64 *)buf;
#endif
}

_int64 xc_int32_out(_int32 val)
{
#ifdef WEIRD
	int j;
	_int64 ret = 0;

	for (j = 3; j >= 0; --j)
	{
		ret += (val & 255) << (8 * (3-j));
		val >>= 8;
	}

	return ret;
#else
	unsigned char buf[8];

	memcpy(buf, &val, 4);
	memset(buf+4, 0, 4);
	return *(_int64 *)buf;
#endif
}

_int64 xc_int64_out(_int64 val)
{
	return val;
}

_int64 xc_float32_out(_float32 val)
{
//#ifdef WEIRD
	int j, sz;
	_int32 e;
	_float32 f;

	f = frexp(val, &e);
	e += mantissa[2];
	sz = mantissa[2];
	for (j = 0; j < sz; ++j)
	{
		f *= 2;
		--e;
	}

	return write_float2((_int64)f, e, 2);
/*#else
	unsigned char buf[8];

	memcpy(buf, &val, 4);
	memset(buf+4, 0, 4);
	return *(_int64 *)buf;
#endif*/
}

_int64 xc_float64_out(_float64 val)
{
//#ifdef WEIRD
	int j, sz;
	_int32 e;
	_float64 f;

	f = frexp(val, &e);
	e += mantissa[3];
	sz = mantissa[3];
	for (j = 0; j < sz; ++j)
	{
		f *= 2;
		--e;
	}

	return write_float2((_int64)f, e, 3);
/*#else
	return *(_int64 *)&val;
#endif*/
}

_int64 _xc_ptr_out(xinst *c_inst_p, void *val)
{
#ifdef WEIRD
	int j;
	_int32 ptr;
	_int64 ret = 0;
#endif

	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return 0;
	}

	if (!val)
		return 0;

	if (val < (void*)c_inst_p->mem || val >= (void*)&c_inst_p->mem[c_inst_p->memsz])
	{
		xc_set_errno(XC_ERR_SEGFAULT);
		return 0;
	}

#ifdef WEIRD
	ptr = (_int32)((int)val - (int)c_inst_p->mem);
	for (j = 3; j >= 0; --j)
	{
		ret += (ptr & 255) << (8 * (3-j));
		ptr >>= 8;
	}

	return ret;
#else
	return (_int64)val - (_int64)c_inst_p->mem;
#endif
}

// Standardized memory access functions

void _xc_mem_write(xinst *c_inst_p, unsigned char *ptr, _int64 val, int sz)
{
	int j;

	for (j = 0; j < sz; ++j)
	{
		*ptr++ = (char)(val & 255);
		val >>= 8;
	}
}

_int64 _xc_mem_read(xinst *c_inst_p, unsigned char *ptr, int sz)
{
	_int64 ret = 0;
	int j;

	for (j = 0; j < sz; ++j)
		ret |= *ptr++ << (8 * j);
	return ret;
}

_int64 xc_func_call(xinst *c_inst_p, int loc, int argc, ...)
{
/*	int j, op;
	va_list va;
	_int64 *arr;
	stackt *s;
	pstack *p, *old;
	_int64 ret;
	xinst *_xinst = c_inst_p;

	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return 0;
	}

	if (loc < 0 || loc > c_inst_p->code_end - c_inst_p->code)
	{
		xc_set_errno(XC_ERR_SEGFAULT);
		return 0;
	}

	arr = calloc(argc, sizeof(_int64));
	va_start(va, argc);
	for (j = 0; j < argc; ++j)
		arr[j] = va_arg(va, _int64);
	va_end(va);

	for (j = argc-1; j >= 0; --j)
	{
		s = (stackt *)calloc(1, sizeof(stackt));
		s->next = c_inst_p->stack;
		c_inst_p->stack = s;
		memcpy(s->n, &arr[j], sizeof(_int64));
	}

	p = (pstack *)calloc(1, sizeof(pstack));
	p->next = old = c_inst_p->cstack;
	c_inst_p->cstack = p;
	c_inst_p->cstack->ptr = c_inst_p->cp;
	c_inst_p->cp = c_inst_p->code + loc;

	op = c_inst_p->priority;
	c_inst_p->priority = 1;
	while (do_xinst(c_inst_p))
		if (c_inst_p->cstack == old)
		{
			ret = xc_mem_read(c_inst_p->stack->n, 8);
			s = c_inst_p->stack;
			c_inst_p->stack = s->next;
			free(s);
			c_inst_p->priority = op;
			return ret;
		}

	c_inst_p->priority = op;*/
	return 0;
}

int _xc_atexit(xinst *c_inst_p, _int32 loc)
{
	lint *p;

	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return 0;
	}

	if (loc < 0 || loc > c_inst_p->code_end - c_inst_p->code)
	{
		xc_set_errno(XC_ERR_SEGFAULT);
		return 0;
	}

	p = (lint *)calloc(1, sizeof(pstack));
	p->next = c_inst_p->atex;
	c_inst_p->atex = p;
	c_inst_p->atex->n = loc;

	return 0;
}

int _xc_set_ivar(xinst *c_inst_p, int n, _int32 val)
{
	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return 0;
	}
	if (n < 0 || n >= XC_NUM_IVARS)
	{
		xc_set_errno(XC_ERR_SEGFAULT);
		return 0;
	}

	c_inst_p->ivars[n] = val;
	return 1;
}

_int32 _xc_get_ivar(xinst *c_inst_p, int n)
{
	if (!c_inst_p)
	{
		xc_set_errno(XC_ERR_NOT_EXECUTING);
		return 0;
	}
	if (n < 0 || n >= XC_NUM_IVARS)
	{
		xc_set_errno(XC_ERR_SEGFAULT);
		return 0;
	}

	return c_inst_p->ivars[n];
}

// Instance void* tags

void *xc_get_tag(xinst *x)
{
	return x->tag;
}

void xc_set_tag(xinst *x, void *t)
{
	x->tag = t;
}

// Ping functions

int xc_ping(xnet *x, time_t to)
{
	unsigned char ch;
	time_t t = time(0);

	if (!xnet_sendc(x, (char)OP_PING))
		return 0;
	while (!xnet_newdata(x))
		if (time(0) - t > to)
			return 0;
	if (!xnet_recv(x, &ch, 1))
		return 0;
	if (ch != OP_PONG)
		return 0;

	return 1;
}
