#include <xc.h>
#include <stdio.h>

#define MEMSZ 66666

xc_func(get_int)
{
	int n;

	printf("Int: ");
	scanf("%d", &n);
	return xc_int32_out(n);
}

xc_func(get_char)
{
	int ch;

	printf("Char :");
	scanf("%c", &ch);
	return ch;
}

xc_func(get_string)
{
	char *str;

	if (!argc)
		return 0;

	errno = 0;
	str = xc_ptr_in(argv[0], 1);
	if (errno)
	{
		puts("Invalid string pointer!");
		return 0;
	}

	printf("String: ");
	gets(str);
	return argv[0];
}

xc_func(get_float)
{
	float n;

	printf("Float: ");
	scanf("%f", &n);
	return xc_float32_out(n);
}

xc_func(get_double)
{
	float n;

	printf("Double: ");
	scanf("%f", &n);
	return xc_float64_out((_float64)n);
}

xc_func(print_int)
{
	if (!argc)
		return 0;

	printf("Int: %d\n", xc_int32_in(argv[0]));
	return 1;
}

xc_func(print_char)
{
	if (!argc)
		return 0;

	printf("Char: %c\n", (char)argv[0]);
	return 1;
}

xc_func(print_string)
{
	char *str;

	if (!argc)
		return 0;

	errno = 0;
	str = xc_string_in(argv[0]);
	if (errno)
	{
		puts("Invalid string pointer!");
		return 0;
	}

	printf("String: %s\n", str);
	return 1;
}

xc_func(_print_float)
{
	if (!argc)
		return 0;

	printf("Float: %g\n", xc_float32_in(argv[0]));
	return 1;
}

xc_func(print_double)
{
	if (!argc)
		return 0;

	printf("Double: %g\n", xc_float64_in(argv[0]));
	return 1;
}

xc_func(_malloc)
{
	if (!argc)
		return 0;

	return xc_ptr_out(xc_malloc(xc_int32_in(argv[0])));
}

xc_func(_free)
{
	void *ptr;

	if (!argc)
		return 0;

	if (!(ptr = xc_ptr_in(argv[0], 1)))
	{
		puts("Bad pointer");
		return 1;
	}

	if (!xc_free(ptr))
		return 0;

	return 1;
}

main(int argc, char *argv[])
{
	int n;
	xc *x;
	xnet xn, xn2;

	xc_export_func(get_int);
	xc_export_func(get_char);
	xc_export_func(get_string);
	xc_export_func(get_float);
	xc_export_func(get_double);
	xc_export_func(print_int);
	xc_export_func(print_char);
	xc_export_func(print_string);
	xc_export_func(_print_float);
	xc_export_func(print_double);
	xc_export_func(_malloc);
	xc_export_func(_free);

	if (argc < 2)
	{
		if (!(x = load_xc("out.xc")))
		{
			puts("out.xc required in current working directory");
			return 1;
		}

		run_xc(x, MEMSZ, NULL, NULL, NULL, NULL, 0);
		printf("Execution returned %d\n", errno);

		unload_xc(x);
	}
	else if (!strcmp(argv[1], "-s"))
	{
		if (!(x = load_xc("out.xc")))
		{
			puts("out.xc required in current working directory");
			return 1;
		}

		xnet_init();
		xnet_create(&xn, XN_TCP_BLOCKING);
		xnet_create(&xn2, XN_TCP_BLOCKING);
		xnet_listen(&xn, 6241, 10);

		while ((n = xnet_accept(&xn, &xn2)) == -1);

		run_xc(x, MEMSZ, NULL, &xn2, &xn2, &xn2, 0);
		printf("Execution returned %d\n", errno);

		xnet_close(&xn);
		xnet_close(&xn2);

		unload_xc(x);
	}
	else
	{
		xnet_init();
		xnet_create(&xn, XN_TCP_BLOCKING);
		xnet_connect(&xn, argv[1], 6241);

		x = new_xc_stream(255);
		run_xc(x, MEMSZ, NULL, NULL, NULL, &xn, 0);
		printf("Execution returned %d\n", errno);

		xnet_close(&xn);

		unload_xc(x);
	}

	return 0;
}